from relevanceai.base import Base
from typing import List

class Search(Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)
        self._init_experiment_helper()

    def vector(
        self,
        dataset_id: str,
        multivector_query: List,
        positive_document_ids: dict = {},
        negative_document_ids: dict = {},
        vector_operation="sum",
        approximation_depth=0,
        sum_fields=True,
        page_size=20,
        page=1,
        similarity_metric="cosine",
        facets=[],
        filters=[],
        min_score=0,
        select_fields=[],
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale = False,
        search_history_id = None
    ):
        """ 
        Allows you to leverage vector similarity search to create a semantic search engine. Powerful features of VecDB vector search:
        
        1. Multivector search that allows you to search with multiple vectors and give each vector a different weight. e.g. Search with a product image vector and text description vector to find the most similar products by what it looks like and what its described to do. You can also give weightings of each vector field towards the search, e.g. image_vector_ weights 100%, whilst description_vector_ 50% \n
            An example of a simple multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": ["name_vector_"], "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": ["image_vector_"], "alias":"image"},
            >>> ]
            
            An example of a weighted multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

            An example of a weighted multivector query with multiple fields for each vector:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6, "description_vector_":0.3}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

        2. Utilise faceted search with vector search. For information on how to apply facets/filters check out datasets.documents.get_where \n
        3. Sum Fields option to adjust whether you want multiple vectors to be combined in the scoring or compared in the scoring. e.g. image_vector_ + text_vector_ or image_vector_ vs text_vector_. \n
            When sum_fields=True:

            - Multi-vector search allows you to obtain search scores by taking the sum of these scores.
            - TextSearchScore + ImageSearchScore = SearchScore
            - We then rank by the new SearchScore, so for searching 1000 documents there will be 1000 search scores and results

            When sum_fields=False:

            - Multi vector search but not summing the score, instead including it in the comparison!
            - TextSearchScore = SearchScore1
            - ImagSearchScore = SearchScore2
            - We then rank by the 2 new SearchScore, so for searching 1000 documents there should be 2000 search scores and results.

        4. Personalization with positive and negative document ids.

            - For more information about the positive and negative document ids to personalize check out services.recommend.vector
        
        For more even more advanced configuration and customisation of vector search, reach out to us at dev@relevance.ai and learn about our new advanced_vector_search.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        """
        return self.make_http_request(
            "/services/search/vector",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "positive_document_ids": positive_document_ids,
                "negative_document_ids": negative_document_ids,
                "vector_operation": vector_operation,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
                "search_history_id": search_history_id
            }
        )

    def hybrid(
        self,
        dataset_id: str,
        multivector_query: List,
        text: str,
        fields: list,
        edit_distance: int = -1,
        ignore_spaces: bool = True,
        traditional_weight: float = 0.075,
        page_size: int = 20,
        page=1,
        similarity_metric="cosine",
        facets=[],
        filters=[],
        min_score=0,
        select_fields=[],
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale = False,
        search_history_id = None
    ):
        """ 
        Combine the best of both traditional keyword faceted search with semantic vector search to create the best search possible. \n

        For information on how to use vector search check out services.search.vector. \n

        For information on how to use traditional keyword faceted search check out services.search.traditional.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        text : string
            Text Search Query (not encoded as vector)
        fields : list
            Text fields to search against
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: float
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        edit_distance: int
            This refers to the amount of letters it takes to reach from 1 string to another string. e.g. band vs bant is a 1 word edit distance. Use -1 if you would like this to be automated.
        ignore_spaces: bool
            Whether to consider cases when there is a space in the word. E.g. Go Pro vs GoPro.
        traditional_weight: int
            Multiplier of traditional search score. A value of 0.025~0.075 is the ideal range
        """
        return self.make_http_request(
            "/services/search/hybrid",
            method="POST",
            parameters={
                "dataset_id": "ecommerce-experiments",
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "text": text,
                "fields": fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
                "search_history_id": search_history_id,
                "edit_distance": edit_distance,
                "ignore_spaces": ignore_spaces,
                "traditional_weight": traditional_weight,
            }
        )

    def semantic(
        self,
        dataset_id: str,
        multivector_query: list,
        fields: list,
        text: str,
        page_size: int = 20,
        page=1,
        similarity_metric="cosine",
        facets=[],
        filters=[],
        min_score=0,
        select_fields=[],
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale = False
    ):
        """ 
        A more automated hybrid search with a few extra things that automatically adjusts some of the key parameters for more automated and good out of the box results. \n

        For information on how to configure semantic search check out services.search.hybrid.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        text : string
            Text Search Query (not encoded as vector)
        fields : list
            Text fields to search against
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        """
        return self.make_http_request(
            "/services/search/semantic",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "text": text,
                "fields": fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
            }
        )

    def diversity(
        self,
        dataset_id: str,
        cluster_vector_field: str,
        n_clusters: int,
        multivector_query: list,
        positive_document_ids: dict = {},
        negative_document_ids: dict = {},
        vector_operation="sum",
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric="cosine",
        facets=[],
        filters=[],
        min_score=0,
        select_fields=[],
        include_vector=False,
        include_count=True,
        asc=False,
        keep_search_history=False,
        hundred_scale: bool = False,
        search_history_id: str = None,
        n_init: int = 5,
        n_iter: int = 10,
        return_as_clusters: bool = False
    ):
        """ 
        This will first perform an advanced search and then cluster the top X (page_size) search results. Results are returned as such: Once you have the clusters:
        
        >>> Cluster 0: [A, B, C]
        >>> Cluster 1: [D, E]
        >>> Cluster 2: [F, G]
        >>> Cluster 3: [H, I]

        (Note, each cluster is ordered by highest to lowest search score.) \n

        This intermediately returns:

        >>> results_batch_1: [A, H, F, D] (ordered by highest search score)
        >>> results_batch_2: [G, E, B, I] (ordered by highest search score)
        >>> results_batch_3: [C]

        This then returns the final results:

        >>> results: [A, H, F, D, G, E, B, I, C]

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        cluster_vector_field: str
            The field to cluster on.
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: str
            Search history ID, only used for storing search histories.
        n_clusters: int
            Number of clusters to be specified.
        n_init: int
            Number of runs to run with different centroid seeds
        n_iter: int
            Number of iterations in each run
        return_as_clusters: bool
            If True, return as clusters as opposed to results list
        """
        return self.make_http_request(
            "/services/search/diversity",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "positive_document_ids": positive_document_ids,
                "negative_document_ids": negative_document_ids,
                "vector_operation": vector_operation,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
                "search_history_id": search_history_id,
                "cluster_vector_field": cluster_vector_field,
                "n_clusters": n_clusters,
                "n_init": n_init,
                "n_iter": n_iter,
                "return_as_clusters": return_as_clusters,
            }
        )

    def traditional(
        self,
        dataset_id: str,
        text: str,
        fields: list = [],
        edit_distance: int = -1,
        ignore_spaces: bool = True,
        page_size: int = 29,
        page: int = 1,
        select_fields: list = [],
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        search_history_id: str = None
    ):
        """ 
        Traditional Faceted Keyword Search with edit distance/fuzzy matching. \n

        For information on how to apply facets/filters check out datasets.documents.get_where. \n

        For information on how to construct the facets section for your search bar check out datasets.facets.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        text : string
            Text Search Query (not encoded as vector)
        fields : list
            Text fields to search against
        edit_distance: int
            This refers to the amount of letters it takes to reach from 1 string to another string. e.g. band vs bant is a 1 word edit distance. Use -1 if you would like this to be automated.
        ignore_spaces: bool
            Whether to consider cases when there is a space in the word. E.g. Go Pro vs GoPro.
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        search_history_id: string
            Search history ID, only used for storing search histories.
        """
        return self.make_http_request(
            "/services/search/traditional",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "text": text,
                "fields": fields,
                "edit_distance": edit_distance,
                "ignore_spaces": ignore_spaces,
                "page_size": page_size,
                "page": page,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "search_history_id": search_history_id,
            }
        )

    def chunk(
        self,
        dataset_id,
        multivector_query,
        chunk_field,
        chunk_scoring="max",
        chunk_page_size: int = 3,
        chunk_page: int = 1,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: list = [],
        filters: list = [],
        min_score: int = None,
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False
    ):

        """ 
        Chunks are data that has been divided into different units. e.g. A paragraph is made of many sentence chunks, a sentence is made of many word chunks, an image frame in a video. By searching through chunks you can pinpoint more specifically where a match is occuring. When creating a chunk in your document use the suffix "chunk" and "chunkvector". An example of a document with chunks:
        
        >>> {
        >>>     "_id" : "123",
        >>>     "title" : "Lorem Ipsum Article",
        >>>     "description" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry. Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book. It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.",
        >>>     "description_vector_" : [1.1, 1.2, 1.3],
        >>>     "description_sentence_chunk_" : [
        >>>         {"sentence_id" : 0, "sentence_chunkvector_" : [0.1, 0.2, 0.3], "sentence" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry."},
        >>>         {"sentence_id" : 1, "sentence_chunkvector_" : [0.4, 0.5, 0.6], "sentence" : "Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book."},
        >>>         {"sentence_id" : 2, "sentence_chunkvector_" : [0.7, 0.8, 0.9], "sentence" : "It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged."},
        >>>     ]
        >>> }
        
        For combining chunk search with other search check out services.search.advanced_chunk.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        """

        return self.make_http_request(
            "/services/search/chunk",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "chunk_field": chunk_field,
                "chunk_scoring": chunk_scoring,
                "chunk_page_size": chunk_page_size,
                "chunk_page": chunk_page,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale
            }
        )

    def multistep_chunk(
        self,
        dataset_id,
        multivector_query,
        first_step_multivector_query,
        chunk_field,
        chunk_scoring="max",
        chunk_page_size: int = 3,
        chunk_page: int = 1,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: list = [],
        filters: list = [],
        min_score: int = None,
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False,
        first_step_page: int = 1,
        first_step_page_size: int = 20
    ):

        """ 
        Multistep chunk search involves a vector search followed by chunk search, used to accelerate chunk searches or to identify context before delving into relevant chunks. e.g. Search against the paragraph vector first then sentence chunkvector after. \n

        For more information about chunk search check out services.search.chunk. \n

        For more information about vector search check out services.search.vector

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        first_step_multivector_query: list
            Query for advance search that allows for multiple vector and field querying.
        first_step_page: int
            Page of the results
        first_step_page_size: int
            Size of each page of results
        """
        return self.make_http_request(
            "/services/search/multistep_chunk",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "multivector_query": multivector_query,
                "chunk_field": chunk_field,
                "chunk_scoring": chunk_scoring,
                "chunk_page_size": chunk_page_size,
                "chunk_page": chunk_page,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
                "first_step_multivector_query": first_step_multivector_query,
                "first_step_page": first_step_page,
                "first_step_page_size": first_step_page_size,
            }
        )

    def advanced_chunk(
        self,
        dataset_ids,
        chunk_search_query: List,
        min_score: int = None,
        page_size: int = 20,
        include_vector: bool = False,
        select_fields: list = []
    ):
        """ 
        A more advanced chunk search to be able to combine vector search and chunk search in many different ways. \n
        Example 1 (Hybrid chunk search):
        >>> chunk_query = {
        >>>     "chunk" : "some.test",
        >>>     "queries" : [
        >>>         {"vector" : vec1, "fields": {"some.test.some_chunkvector_":1}, 
        >>>         "traditional_query" : {"text":"python", "fields" : ["some.test.test_words"], "traditional_weight": 0.3}, 
        >>>         "metric" : "cosine"},
        >>>         {"vector" : vec, "fields": ["some.test.tt.some_other_chunkvector_"], 
        >>>         "traditional_query" : {"text":"jumble", "fields" : ["some.test.test_words"], "traditional_weight": 0.3}, 
        >>>         "metric" : "cosine"},
        >>>     ]
        >>> }

        Example 2 (combines normal vector search with chunk search):
        >>> chunk_query = {
        >>>     "queries" : [
        >>>         {
        >>>             "queries": [
        >>>                 {
        >>>                     "vector": vec1,
        >>>                     "fields": {
        >>>                         "some.test.some_chunkvector_": 0.9
        >>>                     },
        >>>                     "traditional_query": {
        >>>                         "text": "python",
        >>>                         "fields": [
        >>>                             "some.test.test_words"
        >>>                         ],
        >>>                         "traditional_weight": 0.3
        >>>                     },
        >>>                     "metric": "cosine"
        >>>                 }
        >>>             ],
        >>>             "chunk": "some.test",
        >>>         },
        >>>         {
        >>>             "vector" : vec, 
        >>>             "fields": {
        >>>                 ".some_vector_" : 0.1}, 
        >>>                 "metric" : "cosine"
        >>>                 },
        >>>         ]
        >>>     }

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        chunk_search_query : list
            Advanced chunk query
        min_score: int
            Minimum score for similarity metric
        page_size: int
            Size of each page of results
        include_vector: bool
            Include vectors in the search results
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        """
        return self.make_http_request(
            "/services/search/advanced_chunk",
            method="POST",
            parameters={
                "dataset_ids": dataset_ids,
                "chunk_search_query": chunk_search_query,
                "page_size": page_size,
                "min_score": min_score,
                "include_vector": include_vector,
                "select_fields": select_fields,
            }
        )

    def advanced_multistep_chunk(
        self,
        dataset_ids: list,
        first_step_query: list,
        first_step_text: str,
        first_step_fields: list,
        chunk_search_query: list,
        first_step_edit_distance: int = -1,
        first_step_ignore_space: bool = True,
        first_step_traditional_weight: float = 0.075,
        first_step_approximation_depth: int = 0,
        first_step_sum_fields: bool = True,
        first_step_filters: list = [],
        first_step_page_size: int = 50,
        include_count: bool = True,
        min_score: int = 0,
        page_size: int = 20,
        include_vector: bool = False,
        select_fields: list = []
    ):
        """ 
        Performs a vector hybrid search and then an advanced chunk search. Chunk Search allows one to search through chunks inside a document. The major difference between chunk search and normal search in Vector AI is that it relies on the chunkvector field. Chunk Vector Search. Search with a multiple chunkvectors for the most similar documents. Chunk search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set. \n

        Example 1 (Hybrid chunk search):

        >>> chunk_query = {
        >>>     "chunk" : "some.test",
        >>>     "queries" : [
        >>>         {"vector" : vec1, "fields": {"some.test.some_chunkvector_":1}, 
        >>>         "traditional_query" : {"text":"python", "fields" : ["some.test.test_words"], "traditional_weight": 0.3}, 
        >>>         "metric" : "cosine"},
        >>>         {"vector" : vec, "fields": ["some.test.tt.some_other_chunkvector_"], 
        >>>         "traditional_query" : {"text":"jumble", "fields" : ["some.test.test_words"], "traditional_weight": 0.3}, 
        >>>         "metric" : "cosine"},
        >>>     ]
        >>> }

        Example 2 (combines normal vector search with chunk search):
        >>> chunk_query = {
        >>>     "queries" : [
        >>>         {
        >>>             "queries": [
        >>>                 {
        >>>                     "vector": vec1,
        >>>                     "fields": {
        >>>                         "some.test.some_chunkvector_": 0.9
        >>>                     },
        >>>                     "traditional_query": {
        >>>                         "text": "python",
        >>>                         "fields": [
        >>>                             "some.test.test_words"
        >>>                         ],
        >>>                         "traditional_weight": 0.3
        >>>                     },
        >>>                     "metric": "cosine"
        >>>                 }
        >>>             ],
        >>>             "chunk": "some.test",
        >>>         },
        >>>         {
        >>>             "vector" : vec, 
        >>>             "fields": {
        >>>                 ".some_vector_" : 0.1}, 
        >>>                 "metric" : "cosine"
        >>>                 },
        >>>         ]
        >>>     }

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        first_step_query : list
            First step query
        first_step_text : string
            Text search query (not encoded as vector)
        first_step_fields: list
            Text fields to search against
        chunk_search_query: list
            Advanced chunk query
        first_step_edit_distance: int
            This refers to the amount of letters it takes to reach from 1 string to another string. e.g. band vs bant is a 1 word edit distance. Use -1 if you would like this to be automated.
        first_step_ignore_spaces: bool
            Whether to consider cases when there is a space in the word. E.g. Go Pro vs GoPro.
        first_step_traditional_weight: int
            Multiplier of traditional search score. A value of 0.025~0.075 is the ideal range
        first_step_approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        first_step_sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        first_step_filters: list
            Query for filtering the search results
        first_step_page_size: int
            In the first search, you are more interested in the contents
        include_count: bool
            Include the total count of results in the search results
        min_score: int
            Minimum score for similarity metric
        page_size: int
            Size of each page of results
        include_vector: bool
            Include vectors in the search results
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.

        """
        return self.make_http_request(
            "/services/search/advanced_multistep_chunk",
            method="POST",
            parameters={
                "dataset_ids": dataset_ids,
                "first_step_query": first_step_query,
                "first_step_text": first_step_text,
                "first_step_fields": first_step_fields,
                "chunk_search_query": chunk_search_query,
                "first_step_edit_distance": first_step_edit_distance,
                "first_step_ignore_space": first_step_ignore_space,
                "first_step_traditional_weight": first_step_traditional_weight,
                "first_step_approximation_depth": first_step_approximation_depth,
                "first_step_sum_fields": first_step_sum_fields,
                "first_step_filters": first_step_filters,
                "first_step_page_size": first_step_page_size,
                "include_count": include_count,
                "min_score": min_score,
                "page_size": page_size,
                "include_vector": include_vector,
                "select_fields": select_fields,
            }
        )

    def _init_experiment_helper(self, categories = ['chunk', 'vector', 'diversity', 'traditional']):
        self.categories = categories
        self.traditional_search_doc = "https://docs.relevance.ai/docs/better-text-search-with-hybrid"
        self.vector_search_doc = "https://docs.relevance.ai/docs/pure-word-matching-pure-vector-search-or-combination-of-both"
        self.diversity_search_doc = "https://docs.relevance.ai/docs/better-text-search-diversified-search-results"
        self.hybrid_search_doc = "https://docs.relevance.ai/docs/pure-word-matching-pure-vector-search-or-combination-of-both-1"
        self.semantic_search_doc = "https://docs.relevance.ai/docs/pure-word-matching-pure-vector-search-or-combination-of-both-2"
        self.chunk_search_doc =  "https://docs.relevance.ai/docs/better-text-search-chunk-search"
        self.multistep_chunk_doc = "https://docs.relevance.ai/docs/fine-grained-search-search-on-chunks-of-text-data"
        self.advanced_chunk_doc = "https://docs.relevance.ai/docs/fine-grained-search-search-on-chunks-of-text-data-1"
        self.advanced_multistep_chunk_doc = "https://docs.relevance.ai/docs/fine-grained-search-search-on-chunks-of-text-data-2"

        self.initiative_messages = "What else to experiment with :)\n"
        self.category_initiative_messages = {
            'chunk': "if you are searching on large pieces of text, you could chunk your data and try\n",
            'vector': "if you are looking for strong conceptual relations and not just word matching, you could try\n",
            'diversity': "if you are looking for strong conceptual relations as well as diverse results, you could try\n",
            'traditional': "if you are looking for specific text such as id, names, etc., you could try\n",
        }

    def make_suggestion(self):
        if hasattr(self, "_last_used_endpoint"):
            self.last_search = self._last_used_endpoint.split('/')[-1]
        else:
            self.last_search = None
        suggestion = self.initiative_messages
        if 'traditional' in self.categories and self.last_search != 'traditional':
            suggestion += self.category_initiative_messages['traditional']
            suggestion += f"   * traditional search ({self.traditional_search_doc})\n"

        if 'vector' in self.categories:
            suggestion += self.category_initiative_messages['vector']
            if self.last_search != "vector":
                 suggestion += f"   * vector search ({self.vector_search_doc})\n"
            if self.last_search != "hybrid":
                 suggestion += f"   * hybrid search ({self.hybrid_search_doc})\n"
            if self.last_search != "semantic":
                 suggestion += f"   * semantic search ({self.semantic_search_doc})\n"

        if 'diversity' in self.categories and self.last_search != 'diversity':
            suggestion += self.category_initiative_messages['diversity']
            suggestion += f"   * diversity search ({self.diversity_search_doc})\n"

        if 'chunk' in self.categories:
          suggestion += self.category_initiative_messages['chunk']
          if self.last_search != "chunk":
              suggestion += f"   * chunk search ({self.chunk_search_doc})\n"
          if self.last_search != "multistep_chunk":
              suggestion += f"   * multistep_chunk search ({self.multistep_chunk_doc})\n"
          if self.last_search != "advanced_chunk":
              suggestion += f"   * advanced_chunk search ({self.advanced_chunk_doc})\n"
          if self.last_search != "advanced_multistep_chunk":
              suggestion += f"   * advanced_multistep_chunk search ({self.advanced_multistep_chunk_doc})\n"

        return {"search":suggestion}



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import json
import warnings

import plotly.graph_objs as go

from dataclasses import dataclass
from typeguard import typechecked

from relevanceai.api.client import BatchAPIClient
from relevanceai.base import Base
from relevanceai.vector_tools.constants import DIM_REDUCTION, CLUSTER

from relevanceai.vector_tools.cluster import Cluster, ClusterBase
from relevanceai.vector_tools.dim_reduction import DimReduction, DimReductionBase

from typing import List, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal

from doc_utils import DocUtils

RELEVANCEAI_BLUE = "#1854FF"
MARKER_SIZE = 5


@dataclass
class Projector(BatchAPIClient, Base, DocUtils):
    """
    Projector class.

    Example:
        >>> from relevanceai import Client
        >>> project = input()
        >>> api_key = input()
        >>> client = Client(project, api_key)
        >>> client.projector.plot(
                dataset_id, vector_field, number_of_points_to_render, random_state,
                dr, dr_args, dims,
                vector_label, vector_label_char_length,
                color_label, colour_label_char_length,
                hover_label,
                cluster, cluster_args,
                )
    """

    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    @typechecked
    def plot(
        self,
        dataset_id: str,
        vector_field: str,
        number_of_points_to_render: int = 1000,
        random_state: int = 0,
        # Dimensionality reduction args
        dr: Union[DIM_REDUCTION, DimReductionBase] = "pca",
        dr_args: Union[None, Dict] = None,
        # TODO: Add support for 2
        dims: Literal[2, 3] = 3,
        # Plot rendering args
        vector_label: Union[None, str] = None,
        vector_label_char_length: Union[None, int] = 50,
        colour_label: Union[None, str] = None,
        colour_label_char_length: Union[None, int] = 20,
        hover_label: List[str] = [],
        # Cluster args
        cluster: Union[CLUSTER, ClusterBase] = None,
        cluster_args: Union[None, Dict] = None,
        num_clusters: Union[None, int] = 10,
        marker_colour: str = RELEVANCEAI_BLUE,
        marker_size: int = MARKER_SIZE
    ):
        """
        Plot function for Embedding Projector class

        To write your own custom dimensionality reduction, you should inherit from DimReductionBase:
        from relevanceai.visualise.dim_reduction import DimReductionBase
        class CustomDimReduction(DimReductionBase):
            def fit_transform(self, vectors):
                return np.arange(512, 2)

        Example:
            >>> from relevanceai import Client
            >>> project = input()
            >>> api_key = input()
            >>> client = Client(project, api_key)
            >>> client.projector.plot(
                    dataset_id, vector_field, number_of_points_to_render, random_state,
                    dr, dr_args, dims,
                    vector_label, vector_label_char_length,
                    color_label, colour_label_char_length,
                    hover_label,
                    cluster, cluster_args,
                    )
        """
        self.dataset_id = dataset_id
        self.vector_label = vector_label
        self.vector_field = vector_field
        self.number_of_points_to_render = number_of_points_to_render
        self.random_state = random_state
        self.vector_label_char_length = vector_label_char_length
        self.colour_label = colour_label
        self.colour_label_char_length = colour_label_char_length
        self.hover_label = hover_label
        self.cluster = cluster
        self.num_clusters = num_clusters
        self.dr = dr
        self.dr_args = dr_args
        self.dims = dims
        self.cluster_args = cluster_args

        if (vector_label is None) and (colour_label is None):
            warnings.warn(
                f"A vector_label or colour_label has not been specified.")

        if number_of_points_to_render and number_of_points_to_render > 1000:
            warnings.warn(
                f"You are rendering over 1000 points, this may take some time ..."
            )

        number_of_documents = number_of_points_to_render
        self.vector_fields = self._get_vector_fields()

        labels = ["_id", vector_field, vector_label, colour_label]
        if hover_label:
            labels += hover_label
        fields = [label for label in labels if label]
        self.docs = self.get_documents(
            dataset_id, number_of_documents=number_of_documents, batch_size=1000, select_fields=fields
        )
        self._remove_empty_vector_fields(vector_field)

        return self.plot_from_docs(self.docs, self.dims, marker_size, marker_colour)

    def plot_from_docs(self, docs: List[Dict[str, Any]], dims: int, marker_size: int, marker_colour: str, *args, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

        if self._is_valid_vector_name(self.vector_field):

            self.vectors = np.array(
                self.get_field_across_documents(self.vector_field, docs)
            )
            self.vectors_dr = DimReduction.dim_reduce(
                vectors=self.vectors, dr=self.dr, dr_args=self.dr_args, dims=self.dims
            )
            points = {
                "x": self.vectors_dr[:, 0],
                "y": self.vectors_dr[:, 1],
                "_id": self.get_field_across_documents("_id", docs),
            }
            if dims == 3:
                points["z"] = self.vectors_dr[:, 2]

            self.embedding_df = pd.DataFrame(points)
            if self.hover_label and all(
                self._is_valid_label_name(l) for l in self.hover_label
            ):
                self.embedding_df = pd.concat(
                    [self.embedding_df, pd.DataFrame(docs)], axis=1
                )

            if self.vector_label and self._is_valid_label_name(self.vector_label):
                self.labels = self.get_field_across_documents(
                    field=self.vector_label, docs=docs
                )
                self.embedding_df[self.vector_label] = self.labels
                self.embedding_df["labels"] = self.labels

            self.legend = None
            if self.colour_label and self._is_valid_label_name(self.colour_label):
                self.labels = self.get_field_across_documents(
                    field=self.colour_label, docs=docs
                )
                self.embedding_df["labels"] = self.labels
                self.embedding_df[self.colour_label] = self.labels
                self.legend = "labels"

            # TODO: refactor Cluster
            if self.cluster:
                # _cluster = Cluster(
                #     **self.base_args,
                #     vectors=self.vectors,
                #     cluster=self.cluster,
                #     cluster_args=self.cluster_args,
                #     k=self.num_clusters,
                # )
                # self.cluster_labels = _cluster.cluster_labels
                self.cluster_labels = Cluster.cluster(
                    vectors=self.vectors,
                    cluster=self.cluster,
                    cluster_args=self.cluster_args,
                )
                self.embedding_df["cluster_labels"] = self.cluster_labels
                self.legend = "cluster_labels"

            self.embedding_df.index = self.embedding_df["_id"]
            return self._generate_fig(
                embedding_df=self.embedding_df, legend=self.legend, marker_size=marker_size, marker_colour=marker_colour
            )

    def _get_vector_fields(self) -> List[str]:
        """
        Returns list of valid vector fields from dataset schema
        """
        self.schema = self.datasets.schema(self.dataset_id)
        self.vector_dim = self.schema[self.vector_field]["vector"]
        return [k for k in self.schema.keys() if k.endswith("_vector_")]

    def _is_valid_vector_name(self, vector_name: str) -> bool:
        """
        Check vector field name is valid
        """
        if vector_name in self.schema.keys():
            if vector_name in self.vector_fields:
                return True
            else:
                raise ValueError(f"{vector_name} is not a valid vector name")
        else:
            raise ValueError(
                f"{vector_name} is not in the {self.dataset_id} schema")

    def _is_valid_label_name(self, label_name: str) -> bool:
        """
        Check vector label name is valid. Checks that it is either numeric or text
        """
        if label_name == "_id":
            return True
        if label_name in list(self.schema.keys()):
            if self.schema[label_name] in ["numeric", "text"]:
                return True
            else:
                raise ValueError(f"{label_name} is not a valid label name")
        else:
            raise ValueError(
                f"{label_name} is not in the {self.dataset_id} schema")

    def _remove_empty_vector_fields(self, vector_field: str) -> List[Dict]:
        """
        Remove documents with empty vector fields
        """
        self.docs = [d for d in self.docs if d.get(vector_field)]
        return self.docs

    def _generate_fig(
        self,
        embedding_df: pd.DataFrame,
        marker_size: int,
        marker_colour: str,
        legend: Union[None, str],
    ) -> go.Figure:
        """
        Generates the Scatter plot
        """
        plot_title = f"<b>{self.dims}D Embedding Projector Plot<br>Dataset Id: {self.dataset_id} - {len(embedding_df)} points<br>Vector Field: {self.vector_field}<br></b>"
        self.hover_label = ["_id"] + self.hover_label
        text_labels = None
        plot_mode = "markers"

        """
        Generates data for word plot
        If vector_label set, generates text_labels, otherwise shows points only
        """
        if self.vector_label:
            plot_title = plot_title.replace(
                "</b>", f"Vector Label: {self.vector_label}<br></b>"
            )
            plot_mode = "text+markers"
            text_labels = embedding_df["labels"]
            if self.vector_label_char_length and not self.cluster:
                plot_title = plot_title.replace(
                    "<br></b>",
                    f"  Char Length: {self.vector_label_char_length}<br></b>",
                )
                text_labels = embedding_df["labels"].apply(
                    lambda x: x[: self.vector_label_char_length] + "..."
                )

            self.hover_label.insert(1, self.vector_label)

            # self.hover_label = [self.vector_label] + self.hover_label
            # self.hover_label = list(set(self.hover_label))

            # TODO: We can change this later to show top 100 neighbours of a selected word
            #  # Regular displays the full scatter plot with only circles
            # if wordemb_display_mode == 'regular':
            #     plot_mode = 'markers'
            # # Nearest Neighbors displays only the 200 nearest neighbors of the selected_word, in text rather than circles
            # elif wordemb_display_mode == 'neighbors':
            #     if not selected_word:
            #         return go.Figure()
            #     plot_mode = 'text'
            #     # Get the nearest neighbors indices
            #     dataset = data_dict[dataset_name].set_index('0')
            #     selected_vec = dataset.loc[selected_word]

            #     nearest_neighbours = get_nearest_neighbours(
            #                             dataset=dataset,
            #                             selected_vec=selected_vec,
            #                             distance_measure_mode=distance_measure_mode,
            #                             )

            #     neighbors_idx = nearest_neighbours[:100].index
            #     embedding_df =  embedding_df.loc[neighbors_idx]

        custom_data, hovertemplate = self._generate_hover_template(
            df=embedding_df, dims=self.dims
        )

        vector_label_scatter_args = {
            "text": text_labels,
            "textposition": "middle center",
            "showlegend": False,
            "mode": plot_mode,
            "marker": {
                "size": marker_size,
                "color": marker_colour,
                "symbol": "circle",
            },
            "customdata": custom_data,
            "hovertemplate": hovertemplate,
        }

        if self.dims == 3:
            scatter = go.Scatter3d(
                x=embedding_df["x"],
                y=embedding_df["y"],
                z=embedding_df["z"],
                **vector_label_scatter_args,
            )

        else:
            scatter = go.Scatter(
                x=embedding_df["x"], y=embedding_df["y"], **vector_label_scatter_args
            )

        data = [scatter]

        """
        Generates data for colour plot if selected
        """
        if self.colour_label:
            plot_title = plot_title.replace(
                "</b>", f"Colour Label: {self.colour_label}<br></b>"
            )
            if self.colour_label_char_length and not self.cluster:
                plot_title = plot_title.replace(
                    "<br></b>",
                    f"  Char Length: {self.colour_label_char_length}<br></b>",
                )
                colour_labels = embedding_df["labels"].apply(
                    lambda x: x[: self.colour_label_char_length] + "..."
                )
                embedding_df["labels"] = colour_labels

            self.hover_label.insert(1, self.colour_label)

            data = []
            groups = embedding_df.groupby(legend)
            for idx, val in groups:
                # if self.vector_label:
                #     plot_mode = "text+markers"
                #     text_labels = val["labels"]
                #     if self.vector_label_char_length and not self.cluster:
                #         plot_title = plot_title.replace(
                #             "<br></b>",
                #             f"  Char Length: {self.vector_label_char_length}<br></b>",
                #         )
                #         text_labels = val["labels"].apply(
                #             lambda x: x[: self.vector_label_char_length] + "..."
                #         )

                #     self.hover_label = [self.vector_label] + self.hover_label
                #     self.hover_label = list(set(self.hover_label))

                custom_data, hovertemplate = self._generate_hover_template(
                    df=val, dims=self.dims
                )

                colour_label_scatter_args = (
                    {  # text:[ idx for _ in range(val["x"].shape[0]) ],
                        "text": text_labels,
                        "textposition": "top center",
                        "showlegend": False,
                        "mode": plot_mode,
                        "marker": {"size": marker_size, "symbol": "circle"},
                        "customdata": custom_data,
                        "hovertemplate": hovertemplate,
                    }
                )

                if self.dims == 3:
                    scatter = go.Scatter3d(
                        name=idx,
                        x=val["x"],
                        y=val["y"],
                        z=val["z"],
                        **colour_label_scatter_args,
                    )

                else:

                    scatter = go.Scatter(
                        name=idx, x=val["x"], y=val["y"], **colour_label_scatter_args
                    )

                data.append(scatter)

        """
        Generating figure
        """
        axes = {
            "title": "",
            "showgrid": True,
            "zeroline": False,
            "showticklabels": False,
        }
        layout = go.Layout(
            margin={"l": 0, "r": 0, "b": 0, "t": 0},
            scene={"xaxis": axes, "yaxis": axes, "zaxis": axes},
        )

        if self.cluster:
            plot_title = plot_title.replace(
                "</b>",
                f"<b>Cluster Method: {self.cluster}<br>Num Clusters: {self.num_clusters}</b>",
            )
        fig = go.Figure(data=data, layout=layout)
        fig.update_layout(
            title={
                "text": plot_title,
                "y": 0.1,
                "x": 0.1,
                "xanchor": "left",
                "yanchor": "bottom",
                "font": {"size": 10},
            },
        )
        if legend and self.colour_label:
            fig.update_layout(
                legend={
                    "title": {"text": self.colour_label, "font": {"size": 12}},
                    "font": {"size": 10},
                    "itemwidth": 30,
                    "tracegroupgap": 1,
                }
            )
        return fig

    def _generate_hover_template(
        self, df: pd.DataFrame, dims: int
    ) -> Tuple[Optional[pd.DataFrame], Optional[str]]:
        """
        Generating hover template
        """
        self.hover_label = list(sorted(set(self.hover_label)))
        custom_data = df[self.hover_label]
        custom_data_hover = [
            f"{c}: %{{customdata[{i}]}}"
            for i, c in enumerate(self.hover_label)
            if self._is_valid_label_name(c)
        ]

        if dims == 2:
            coord_info = "X: %{x}   Y: %{y}"
        else:
            coord_info = "X: %{x}   Y: %{y}   Z: %{z}"

        hovertemplate = (
            "<br>".join(
                [
                    coord_info,
                ]
                + custom_data_hover
            )
            + "<extra></extra>"
        )

        return custom_data, hovertemplate



# -*- coding: utf-8 -*-

from abc import abstractmethod
import pandas as pd
import numpy as np
import json

from doc_utils.doc_utils import DocUtils

from typing import List, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal

from relevanceai.base import Base
from relevanceai.logger import LoguruLogger
from relevanceai.vector_tools.constants import DIM_REDUCTION, DIM_REDUCTION_DEFAULT_ARGS


class DimReductionBase(LoguruLogger):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    # @abstractmethod
    def fit_transform(self, 
            vectors: np.ndarray, 
            dr_args: Dict[Any, Any], 
            dims: int
    ) -> np.ndarray:
        raise NotImplementedError



class PCA(DimReductionBase):
    def fit_transform(self, 
        vectors: np.ndarray, 
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS['pca'], 
        dims: int = 3
    ) -> np.ndarray:
        from sklearn.decomposition import PCA
        self.logger.debug(f"{dr_args}")
        pca = PCA(n_components=min(dims, vectors.shape[1]), **dr_args)
        return pca.fit_transform(vectors)


class TSNE(DimReductionBase):
    def fit_transform(self, 
        vectors: np.ndarray, 
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS['tsne'], 
        dims: int = 3
    ) -> np.ndarray:
        from sklearn.decomposition import PCA
        from sklearn.manifold import TSNE
        pca = PCA(n_components=min(10, vectors.shape[1]))
        data_pca = pca.fit_transform(vectors)
        self.logger.debug(f"{dr_args}")
        tsne = TSNE(n_components=dims, **dr_args)
        return tsne.fit_transform(data_pca)


class UMAP(DimReductionBase):
    def fit_transform(self, 
        vectors: np.ndarray, 
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS['umap'], 
        dims: int = 3 
    ) -> np.ndarray:
        try:
            from umap import UMAP
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall umap\n \
                pip install -U relevanceai[umap]"
            )
        self.logger.debug(f"{dr_args}")
        umap = UMAP(n_components=dims, **dr_args)
        return umap.fit_transform(vectors)


class Ivis(DimReductionBase):
    def fit_transform(self, 
        vectors: np.ndarray, 
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS['tsne'], 
        dims: int = 3
    ) -> np.ndarray:
        try:
            from ivis import Ivis
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall ivis\n \
                CPU: pip install -U relevanceai[ivis-cpu]\n \
                GPU: pip install -U relevanceai[ivis-gpu]"
            )
        self.logger.debug(f"{dr_args}")
        ivis = Ivis(embedding_dims=dims, **dr_args)
        if ivis.batch_size > vectors.shape[0]:
            ivis.batch_size = vectors.shape[0]
        vectors_dr = ivis.fit(vectors).transform(vectors)
        return vectors_dr


class DimReduction(Base, DimReductionBase):

    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    @staticmethod
    def dim_reduce(
        vectors: np.ndarray,
        dr: Union[DIM_REDUCTION, DimReductionBase],
        dr_args: Union[None, dict],
        dims: Literal[2, 3],
    ) -> np.ndarray:
        """
        Dimensionality reduction
        """
        if isinstance(dr, str):
            if dr_args is None:
                dr_args = DIM_REDUCTION_DEFAULT_ARGS[dr]
            if dr == "pca":
                return PCA().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "tsne":
                return TSNE().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims) 
            elif dr == "umap":
                return UMAP().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "ivis":
                return Ivis().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)

        elif isinstance(dr, DimReductionBase):
            return dr().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims) 


"""API Client
"""
from relevanceai.base import Base
from relevanceai.api.endpoints.admin import Admin
from relevanceai.api.endpoints.datasets import Datasets
from relevanceai.api.endpoints.services import Services
from relevanceai.datasets import ExampleDatasets


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class APIClient(Base):
    """API Client"""

    def __init__(self, project: str, api_key: str):
        self.datasets = Datasets(project=project, api_key=api_key)
        self.services = Services(project=project, api_key=api_key)
        self.example_datasets = ExampleDatasets()
        self.admin = Admin(project=project, api_key=api_key)
        super().__init__(project, api_key)


# -*- coding: utf-8 -*-

from abc import abstractmethod
import pandas as pd
import numpy as np
import json
import warnings

from dataclasses import dataclass

from typing import List, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal

from doc_utils import DocUtils

from relevanceai.base import Base
from relevanceai.logger import LoguruLogger
from relevanceai.vector_tools.constants import CLUSTER, CLUSTER_DEFAULT_ARGS


class ClusterBase(LoguruLogger, DocUtils):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        """Return the 
        """
        raise NotImplementedError
    
    def fit_documents(
        self,
        vector_field: list,
        docs: list,
        alias: str="default",
        cluster_field: str="_clusters_",
        return_only_clusters: bool=True
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.
        
        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        docs: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document

        """
        if len(vector_field) == 1:
            vectors = self.get_field_across_documents(vector_field[0], docs)
        else:
            raise ValueError("We currently do not support more than 1 vector field yet. This will be supported in the future.")
        cluster_labels = self.fit_transform(vectors)
        # Label the clusters
        cluster_labels = self._label_clusters(cluster_labels)
        self.set_field_across_documents(
            f"{cluster_field}.{vector_field}.{alias}", cluster_labels, docs
        )
        if return_only_clusters:
            return [{"_id": d.get("_id"), cluster_field: d.get(cluster_field)} for d in docs]
        return docs

    def to_metadata(self):
        """You can also store the metadata of this clustering algorithm
        """
        raise NotImplementedError
    
    def _label_cluster(self, label: Union[int, str]):
        if isinstance(label, (int, float)):
            return "cluster_" + str(label)
        return str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]

class CentroidCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        raise NotImplementedError
    
    @abstractmethod
    def get_centers(self) -> Union[np.ndarray, List[list]]:
        """Get centers for the centroid-based clusters
        """
        raise NotImplementedError
    
    def get_centroid_docs(self) -> List:
        """Get the centroid documents to store.
        """
        self.centers = self.get_centers()
        if isinstance(self.centers, np.ndarray):
            self.centers = self.centers.tolist()
        return [
            {
                "_id": f"cluster_{i}",
                "centroid_vector_": self.centers[i]
            } for i in range(len(self.centers))
        ]

class DensityCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    def fit_transform(self, vectors):
        raise NotImplementedError


class KMeans(CentroidCluster):
    def __init__(
        self,
        k: Union[None, int] = 10,
        init: str = "k-means++",
        verbose: bool = True,
        compute_labels: bool = True,
        max_no_improvement: int=2
     ):
        """
        Kmeans Centroid Clustering

        Parameters
        ------------
        k: int
            The number of clusters
        init: str
            The optional parameter to be clustering
        verbose: bool
            If True, will print what is happening 
        compute_labels: bool
            If True, computes the labels of the cluster 
        max_no_improvement: int
            The maximum number of improvemnets
        """
        self.k = k
        self.init = init
        self.verbose = verbose
        self.compute_labels = compute_labels
        self.max_no_improvement = max_no_improvement

    def _init_model(self):
        from sklearn.cluster import MiniBatchKMeans
        self.km = MiniBatchKMeans(
            n_clusters=self.k, 
            init=self.init,
            verbose=self.verbose,
            compute_labels=self.compute_labels,
            max_no_improvement=self.max_no_improvement
        )
        return

    def fit_transform(
        self,
        vectors: Union[np.ndarray, List]
    ):
        """
        Fit and transform transform the vectors
        """
        if not hasattr(self, "km"):
            self._init_model()
        self.km.fit(vectors)
        cluster_labels = self.km.labels_.tolist()
        # cluster_centroids = km.cluster_centers_
        return cluster_labels

    def get_centers(self):
        """Returns centroids of clusters
        """
        return [list(i) for i in self.km.cluster_centers_]

    def to_metadata(self):
        """Editing the metadata of the function
        """
        return {
            "k": self.k,
            "init": self.init,
            "verbose": self.verbose,
            "compute_labels": self.compute_labels,
            "max_no_improvement": self.max_no_improvement
        }

# class KMedoids(CentroidCluster):
#     def fit_transform(self, 
#         vectors: np.ndarray, 
#         cluster_args: Optional[Dict[Any, Any]] = CLUSTER_DEFAULT_ARGS['kmedoids'], 
#         k: Union[None, int] = 10,
#     ) -> np.ndarray:
#         try:
#             from sklearn_extra.cluster import KMedoids
#         except ModuleNotFoundError as e:
#             raise ModuleNotFoundError(
#                 f"{e}\nInstall umap\n \
#                 pip install -U relevanceai[kmedoids]"
#             )
#         self.logger.debug(f"{cluster_args}")
#         km = KMedoids(n_clusters=k, **cluster_args).fit(vectors)
#         cluster_labels = km.labels_
#         # cluster_centroids = km.cluster_centers_
#         return cluster_labels


class HDBSCAN(DensityCluster):
    def fit_transform(self, 
        vectors: np.ndarray, 
        cluster_args: Optional[Dict[Any, Any]] = CLUSTER_DEFAULT_ARGS['hdbscan'], 
        min_cluster_size: Union[None, int] = 10,
    ) -> np.ndarray:
        try:
            from hdbscan import HDBSCAN
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall hdbscan\n \
                pip install -U relevanceai[hdbscan]"
            )
        self.logger.debug(f"{cluster_args}")
        hdbscan = HDBSCAN(min_cluster_size=min_cluster_size, **cluster_args).fit(vectors)
        cluster_labels = hdbscan.labels_
  
        return cluster_labels


class Cluster(Base, ClusterBase):

    def __init__(self, project, api_key):
        self.project = project
        super().__init__(project, api_key)

    @staticmethod
    def _choose_k(vectors: np.ndarray):
        """ "
        Choose k clusters
        """
        # Partitioning methods
        # if check_type(self.cluster, CLUSTER_NUMERIC):
        """
        Scaled_inertia = inertia(k)/inertia(k=1) + (a * K)
        where a is penalty factor of num_clusters
        """
        warnings.warn("This method is not implemented yet k=10")
        return 10


    @staticmethod
    def cluster(
        vectors: np.ndarray,
        cluster: Union[CLUSTER, ClusterBase],
        cluster_args: Union[None, dict],
        k: Union[None, int] = None,
    ) -> np.ndarray:
        """
        Cluster vectors
        """
        if isinstance(cluster, str):
            if cluster_args is None:
                cluster_args = CLUSTER_DEFAULT_ARGS[cluster]
            if cluster in ['kmeans', 'kmedoids']:
                if (k is None and cluster_args is None) \
                    or ("n_clusters" not in cluster_args.keys()):
                    k = Cluster._choose_k(vectors)
                if cluster == "kmeans":
                    return KMeans(**cluster_args).fit_transform(vectors=vectors)
                elif cluster == "kmedoids":
                    raise NotImplementedError
                    # return KMedioids().fit_transform(vectors=vectors, cluster_args=cluster_args)
            elif cluster == "hdbscan":
                return HDBSCAN().fit_transform(vectors=vectors, cluster_args=cluster_args)
            
        elif isinstance(cluster, ClusterBase):
            return cluster().fit_transform(vectors=vectors, cluster_args=cluster_args)
        raise ValueError("Not valid cluster input.")



"""
All admin-related tasks.
"""
from relevanceai.base import Base


class Admin(Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    def request_read_api_key(self, read_username: str):
        """Creates a read only key for your project. Make sure to save the api key somewhere safe. When doing a search the admin username should still be used.

        Parameters
        ----------
        read_username:
            Read-only project
        """
        return self.make_http_request(
            "/admin/request_read_api_key",
            method="POST",
            parameters={"read_username": read_username},
        )

    def copy_foreign_dataset(
        self,
        dataset_id: str,
        source_dataset_id: str,
        source_project: str,
        source_api_key: str,
        project: str=None,
        api_key: str=None,
    ):
        """Copy a dataset from another user's projects into your project.

        Parameters
        ----------
        dataset_id:
            The dataset to copy
        source_dataset_id:
            The original dataset
        source_project:
            The original project to copy from
        source_api_key:
            The original API key of the project
        project:
            The original project
        api_key:
            The original API key
        """
        return self.make_http_request(
            "/admin/copy_foreign_dataset",
            method="POST",
            parameters={
                "project": self.project if project is None else project,
                "api_key": self.api_key if api_key is None else api_key,
                "dataset_id": dataset_id,
                "source_dataset_id": source_dataset_id,
                "source_project": source_project,
                "source_api_key": source_api_key,
                # "filters": filters
            },
        )



"""
Datasets to mock
"""

from typing import List, Union, Dict, Any
import sys
import pandas as pd
import requests

THIS_MODULE = sys.modules[__name__]
DATASETS = ['games', 'ecommerce_1', 'ecommerce_2', 'ecommerce_3',
            'online_retail', 'news', 'flipkart', 'realestate']


def select_fields_from_json(json, select_fields):
    return [{key: i[key] for key in select_fields} for i in json]

class ExampleDatasets:
    def __init__(self):
        self.datasets = DATASETS

    def list_datasets(self):
        """List of example datasets available to download"""
        return self.datasets

    def get_dataset(self, name, number_of_documents=None, select_fields=[]):
        """
        Download an example dataset
        Parameters
        ----------
        name: string
            Name of example dataset
        number_of_documents: int
            Number of documents to download
        select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
        """
        if name in self.datasets:
            return getattr(THIS_MODULE, f'get_{name}_dataset')(number_of_documents, select_fields)
        else:
            raise ValueError('Not a valid dataset')

    @staticmethod
    def _get_dummy_dataset(db_name, number_of_documents, select_fields=[]):
        from .http_client import Client
        project = "dummy-collections"
        api_key = (
            "UzdYRktIY0JxNmlvb1NpOFNsenU6VGdTU0s4UjhUR0NsaDdnQTVwUkpKZw"  # read access
        )
        client = Client(
            project,
            api_key,
        )
        response = client.datasets.documents.get_where(
            db_name,
            page_size=number_of_documents,
            select_fields=select_fields
        )
        if "message" in response:
            import warnings
            warnings.warn(response["message"])
        docs = response["documents"]
        return docs

    @staticmethod
    def _get_online_dataset(url, number_of_documents, select_fields=[], encoding=None, csv=True):
        if csv:
            data = pd.read_csv(url, index_col=0, encoding=encoding).to_dict(
                orient="records")
        else:
            data = pd.read_excel(url, index_col=0).to_dict(orient="records")
        if number_of_documents:
            data = data[:number_of_documents]
        if len(select_fields) > 0:
            data = select_fields_from_json(data, select_fields)
        return data

    @staticmethod
    def _get_api_dataset(url, number_of_documents, select_fields=[]):
        data = requests.get(url).json()
        if number_of_documents:
            data = data[:number_of_documents]
        if len(select_fields) > 0:
            data = select_fields_from_json(data, select_fields)
        return data


def get_games_dataset(number_of_documents: Union[None, int] = 365, select_fields: list = []) -> List:
    """
    Download an example games dataset (https://www.freetogame.com/) \n
    Total Len: 365 \n
    Sample document:

    >>> {'id': 1,
    >>> 'title': 'Dauntless',
    >>> 'thumbnail': 'https://www.freetogame.com/g/1/thumbnail.jpg',
    >>> 'short_description': 'A free-to-play, co-op action RPG with gameplay similar to Monster Hunter.',
    >>> 'game_url': 'https://www.freetogame.com/open/dauntless',
    >>> 'genre': 'MMORPG',
    >>> 'platform': 'PC (Windows)',
    >>> 'publisher': 'Phoenix Labs',
    >>> 'developer': 'Phoenix Labs, Iron Galaxy',
    >>> 'release_date': '2019-05-21',
    >>> 'freetogame_profile_url': 'https://www.freetogame.com/dauntless'
    >>> }

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 365

    return ExampleDatasets._get_api_dataset("https://www.freetogame.com/api/games", number_of_documents, select_fields)


def get_ecommerce_1_dataset(
    number_of_documents: int = 1000, 
    select_fields: list = []
) -> List[Dict[Any, Any]]:
    """
    Download an example e-commerce dataset \n
    Total Len: 14058 \n 
    Sample document:
    
    >>> {'_id': 'b7fc9acbc9ddd18855f96863d37a4fe9',
    >>> 'uniq_id': 'b7fc9acbc9ddd18855f96863d37a4fe9',
    >>> 'crawl_timestamp': '2016-04-24 18:34:50 +0000',
    >>> 'product_url': 'http://www.flipkart.com/babeezworld-baby-boy-s-romper/p/itmehyhguebbzb6h?pid=DRPEHYHGNAF5UYUQ',
    >>> 'product_name': "Babeezworld Baby Boy's Romper",
    >>> 'product_category_tree': '["Baby Care >> Infant Wear >> Baby Boys\' Clothes >> Dungarees & Jumpsuits >> Dungarees >> Babeezworld Dungarees >> Babeezworld Baby Boy\'s Romper"]',
    >>> 'pid': 'DRPEHYHGNAF5UYUQ',
    >>> 'retail_price': 999,
    >>> 'discounted_price': 499,
    >>> 'image': '["http://img5a.flixcart.com/image/dungaree-romper/x/f/r/1012blue-yellow-3-6-babeezworld-3-6-months-original-imaehydgqkkadjud.jpeg", "http://img5a.flixcart.com/image/dungaree-romper/x/f/r/1012blue-yellow-3-6-babeezworld-3-6-months-original-imaehbf3h3jsmzhb.jpeg", "http://img5a.flixcart.com/image/dungaree-romper/x/f/r/1012blue-yellow-3-6-babeezworld-3-6-months-original-imaehbf4nupsmhzt.jpeg", "http://img5a.flixcart.com/image/dungaree-romper/z/n/u/1012blue-yellow-6-9-babeezworld-6-9-months-original-imaehbfbkynega2z.jpeg", "http://img6a.flixcart.com/image/dungaree-romper/z/n/u/1012blue-yellow-6-9-babeezworld-6-9-months-original-imaehbfbjjffht4e.jpeg"]',
    >>> 'is_FK_Advantage_product': False,
    >>> 'description': "Key Features of Babeezworld Baby Boy's Romper Fabric: Cotton Brand Color: Blue:Yellow,Babeezworld Baby Boy's Romper Price: Rs. 499 Babeezworld presents a cute baby dungaree set for your little one.This dungaree set comes with a trendy round neck cotton t-shirt with shoulder loops for ease and comfort fit. Set is made of soft cotton material and has adjustable straps with two button closures. The front has a special character which gives it a stylish look. This dungaree set is an ideal pick for this summer and is available in multiple colors.,Specifications of Babeezworld Baby Boy's Romper Top Details Sleeve Half Sleeve Number of Contents in Sales Package Pack of 2 Fabric Cotton Type Romper Neck Round Neck General Details Pattern Printed Ideal For Baby Boy's Fabric Care Wash with Similar Colors, Use Detergent for Colors",
    >>> 'product_rating': 'No rating available',
    >>> 'overall_rating': 'No rating available',
    >>> 'brand': 'Babeezworld',
    >>> 'product_specifications': '{"product_specification"=>[{"key"=>"Sleeve", "value"=>"Half Sleeve"}, {"key"=>"Number of Contents in Sales Package", "value"=>"Pack of 2"}, {"key"=>"Fabric", "value"=>"Cotton"}, {"key"=>"Type", "value"=>"Romper"}, {"key"=>"Neck", "value"=>"Round Neck"}, {"key"=>"Pattern", "value"=>"Printed"}, {"key"=>"Ideal For", "value"=>"Baby Boy\'s"}, {"value"=>"Wash with Similar Colors, Use Detergent for Colors"}]}',
    >>> 'image_first': 'http://img5a.flixcart.com/image/dungaree-romper/x/f/r/1012blue-yellow-3-6-babeezworld-3-6-months-original-imaehydgqkkadjud.jpeg',
    >>> 'category': 'Baby Care ',
    >>> 'insert_date_': '2021-08-13T11:38:52.110Z' 
    >>>  ... 
    >>>  }

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 1000
    return ExampleDatasets._get_dummy_dataset("ecommerce-5", number_of_documents, select_fields)


def get_ecommerce_2_dataset(
    number_of_documents: int = 1000, select_fields: list = ["_id", "product_image", "product_link", "product_title", "product_price", "query", "source"]
):
    """
    Download an example e-commerce dataset \n
    Total Len: 739 \n
    Sample document:

    >>> {'_id': '711160239',
    >>> '_unit_id': 711160239,
    >>> 'relevance': 3.67,
    >>> 'relevance:variance': 0.47100000000000003,
    >>> 'product_image': 'https://thumbs4.ebaystatic.com/d/l225/pict/321567405391_1.jpg',
    >>> 'product_link': 'https://www.ebay.com/itm/20-36-Mens-Silver-Stainless-Steel-Braided-Wheat-Chain-Necklace-Jewelry-3-4-5-6MM-/321567405391?pt=LH_DefaultDomain_0&var=&hash=item4adee9354f',
    >>> 'product_price': '$7.99 to $12.99',
    >>> 'product_title': '20-36Mens Silver Stainless Steel Braided Wheat Chain Necklace Jewelry 3/4/5/6MM"',
    >>> 'query': 'steel necklace',
    >>> 'rank': 23,
    >>> 'source': 'eBay',
    >>> 'url': 'https://www.ebay.com/sch/i.html?_from=R40&_trksid=p2050601.m570.l1313.TR11.TRC1.A0.H0.Xplant.TRS0&_nkw=steel%20necklace',
    >>> 'product_description': 'eBay item number:321567405391\n\n\n\tSeller assumes all responsibility for this listing 
    >>> ...
    >>> }

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 1000
    docs = ExampleDatasets._get_dummy_dataset(
        "quickstart_data_sample", number_of_documents, select_fields)
    for d in docs:
        if "image_first" in d:
            d["image"] = d.pop("image_first")
    return docs


def get_online_retail_dataset(number_of_documents: Union[None, int] = 1000, select_fields: list = []) -> List:
    """
    Download an example online retail dataset from UCI machine learning \n
    Total Len: 541909 \n

    Sample document:

    >>> {'Country': 'United Kingdom',
    >>> 'CustomerID': 17850.0,
    >>> 'Description': 'WHITE HANGING HEART T-LIGHT HOLDER',
    >>> 'InvoiceDate': Timestamp('2010-12-01 08:26:00'),
    >>> 'InvoiceNo': 536365,
    >>> 'Quantity': 6,
    >>> 'StockCode': '85123A',
    >>> 'UnitPrice': 2.55}

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 1000
    return ExampleDatasets._get_online_dataset("https://archive.ics.uci.edu/ml/machine-learning-databases/00352/Online%20Retail.xlsx", number_of_documents, select_fields, csv=False)


def get_news_dataset(number_of_documents: Union[None, int] = 250, select_fields: list = []) -> List:
    """
    Download an example news dataset \n
    Total Len: 250 \n
    Sample document:

    >>> {'authors': 'Ruth Harris',
    >>> 'content': 'Sometimes the power of Christmas will make you do wild and wonderful things. You do not need to believe in the Holy Trinity to believe in the positive power of doing good for others. 
    >>> 'domain': 'awm.com',
    >>> 'id': 141,
    >>> 'inserted_at': '2018-02-02 01:19:41.756632',
    >>> 'keywords': nan,
    >>> 'meta_description': nan,
    >>> 'meta_keywords': "['']",
    >>> 'scraped_at': '2018-01-25 16:17:44.789555',
    >>> 'summary': nan,
    >>> 'tags': nan,
    >>> 'title': 'Church Congregation Brings Gift to Waitresses Working on Christmas Eve, Has Them Crying (video)',
    >>> 'type': 'unreliable',
    >>> 'updated_at': '2018-02-02 01:19:41.756664',
    >>> 'url': 'http://awm.com/church-congregation-brings-gift-to-waitresses-working-on-christmas-eve-has-them-crying-video/'}

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 250
    return ExampleDatasets._get_online_dataset("https://raw.githubusercontent.com/several27/FakeNewsCorpus/master/news_sample.csv", number_of_documents, select_fields)


def get_ecommerce_3_dataset(number_of_documents: Union[None, int] = 1000, select_fields: list =[]) -> List:
    """
    Download an example ecommerce dataset (https://data.world/crowdflower/ecommerce-search-relevance) \n
    Total Len: 15528 \n
    Sample document:

    >>> {'_unit_id': 711158459,
    >>> 'product_description': 'The PlayStation 4 system opens the door to an '
    >>>                     'incredible journey through immersive new gaming '
    >>>                     'worlds and a deeply connected gaming community. Step '
    >>>                     'into living, breathing worlds where you are hero of '
    >>>                     '...',
    >>> 'product_image': 'http://thumbs2.ebaystatic.com/d/l225/m/mzvzEUIknaQclZ801YCY1ew.jpg',
    >>> 'product_link': 'http://www.ebay.com/itm/Sony-PlayStation-4-PS4-Latest-Model-500-GB-Jet-Black-Console-/321459436277?pt=LH_DefaultDomain_0&hash=item4ad879baf5',
    >>> 'product_price': '$329.98 ',
    >>> 'product_title': 'Sony PlayStation 4 (PS4) (Latest Model)- 500 GB Jet Black '
    >>>                 'Console',
    >>> 'query': 'playstation 4',
    >>> 'rank': 1,
    >>> 'relevance': 3.67,
    >>> 'relevance:variance': 0.471,
    >>> 'source': 'eBay',
    >>> 'url': 'http://www.ebay.com/sch/i.html?_from=R40&_trksid=p2050601.m570.l1313.TR11.TRC1.A0.H0.Xplant.TRS0&_nkw=playstation%204'}

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 1000
    df = ExampleDatasets._get_online_dataset(
        "https://query.data.world/s/glc7oe2ssd252scha53mu7dy2e7cft", number_of_documents, select_fields, encoding="ISO-8859-1")
    df = pd.DataFrame(df)
    if "product_image" in df.columns:
        df["product_image"] = df["product_image"].str.replace(
            "http://", "https://")
    if "product_link" in df.columns:
        df["product_link"] = df["product_link"].str.replace(
            "http://", "https://")
    if "url" in df.columns:
        df["url"] = df["url"].str.replace("http://", "https://")
    if "_unit_id" in df.columns:
        df["_id"] = df["_unit_id"].astype(str)
    docs = [
        {k: v for k, v in doc.items() if not pd.isna(v)}
        for doc in df.to_dict(orient="records")
    ]
    return docs


def get_flipkart_dataset(
    number_of_documents: Union[None, int] = 19920, 
    select_fields: list =[]
    ) -> List:
    """
    Download an example flipkart ecommerce dataset \n
    Total Len: 19920 \n 
    Sample document:

    >>> {'_id': 0,
    >>> 'product_name': "Alisha Solid Women's Cycling Shorts",
    >>> 'description': "Key Features of Alisha Solid Women's Cycling Shorts Cotton Lycra Navy, Red, Navy,Specifications of Alisha Solid Women's Cycling Shorts Shorts Details Number of Contents in Sales Package Pack of 3 Fabric Cotton Lycra Type Cycling Shorts General Details Pattern Solid Ideal For Women's Fabric Care Gentle Machine Wash in Lukewarm Water, Do Not Bleach Additional Details Style Code ALTHT_3P_21 In the Box 3 shorts",
    >>> 'retail_price': 999.0}

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 19920
    return ExampleDatasets._get_online_dataset("https://raw.githubusercontent.com/arditoibryan/Projects/master/20211108_flipkart_df/flipkart.csv", number_of_documents, select_fields)


def get_realestate_dataset(
    number_of_documents: int = 50,
    select_fields: list =[]
):
    """
    Download an example real-estate dataset \n
    Total Len: 5885 \n
    Sample document:

    >>> {'propertyDetails': {'area': 'North Shore - Lower',
    >>> 'carspaces': 1,
    >>> 'streetNumber': '28',
    >>> 'latitude': -33.8115768,
    >>> 'allPropertyTypes': ['ApartmentUnitFlat'],
    >>> 'postcode': '2066',
    >>> 'unitNumber': '6',
    >>> 'bathrooms': 1.0,
    >>> 'bedrooms': 1.0,
    >>> 'features': ['BuiltInWardrobes', 'InternalLaundry','Intercom', 'Dishwasher'],
    >>> 'street': 'Epping Road',
    >>> 'propertyType': 'ApartmentUnitFlat',
    >>> 'suburb': 'LANE COVE',
    >>> 'state': 'NSW',
    >>> 'region': 'Sydney Region',
    >>> 'displayableAddress': '6/28 Epping Road, Lane Cove',
    >>> 'longitude': 151.166611},
    >>> 'listingSlug': '6-28-epping-road-lane-cove-nsw-2066-14688794',
    >>> 'id': 14688794,
    >>> 'headline': 'Extra large one bedroom unit',
    >>> 'summaryDescription': '<b></b><br />This modern and spacious one-bedroom apartment situated on the top floor, the quiet rear side of a small 2 story boutique block, enjoys a wonderfully private, leafy, and greenly outlook from 2 sides and balcony. A short stroll to city buse...',
    >>> 'advertiser': 'Ray White Lane Cove',
    >>> 'image_url': 'https://bucket-api.domain.com.au/v1/bucket/image/14688794_1_1_201203_101135-w1600-h1065',
    >>> 'insert_date_': '2021-03-01T14:19:22.805086',
    >>> 'labels': [],
    >>> 'image_url_5': 'https://bucket-api.domain.com.au/v1/bucket/image/14688794_5_1_201203_101135-w1600-h1067',
    >>> 'image_url_4': 'https://bucket-api.domain.com.au/v1/bucket/image/14688794_4_1_201203_101135-w1600-h1067',
    >>> 'priceDetails': {'displayPrice': 'Deposit Taken ! Inspection Cancelled thank you !!!'}
    >>> ...
    >>> }

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 50
    return ExampleDatasets._get_dummy_dataset('realestate', number_of_documents, select_fields)

def get_mission_statements_dataset(
    number_of_documents: Union[None, int] = 1433,
    select_fields: list =[]
) -> List:
    """Function to download a sample company mission statement dataset.
    Total Len: 1433
    Sample document:
    {'_id': 0,
    'company': 'Starbucks',
    'text': 'Establish Starbucks as the premier purveyor of the finest coffee in the world while maintaining our uncompromising principles while we grow.'},
    """

    """
    Download an example ompany mission statement dataset \n
    Total Len: 1433 \n 
    Sample document:

    >>> {'_id': 0,
    >>> 'company': 'Starbucks',
    >>> 'text': 'Establish Starbucks as the premier purveyor of the finest coffee in the world while maintaining our uncompromising principles while we grow.'
    >>> },
    

    Parameters
    ----------
    number_of_documents: int
        Number of documents to download
    select_fields : list
            Fields to include in the dataset, empty array/list means all fields.
    """
    if number_of_documents is None:
        number_of_documents = 514330
    return ExampleDatasets._get_online_dataset("https://raw.githubusercontent.com/arditoibryan/Projects/master/20211111_company_statements/companies_preprocessed.csv", number_of_documents, select_fields)


def get_machine_learning_research_dataset():
    """Here we get our Machine Learning research dataset."""
    raise NotImplementedError

get_dummy_ecommerce_dataset = get_ecommerce_1_dataset
get_sample_ecommerce_dataset = get_ecommerce_2_dataset
get_ecommerce_dataset = get_ecommerce_3_dataset



"""access the client via this class
"""
import getpass
import os
from typing import Optional

from doc_utils.doc_utils import DocUtils

from relevanceai.api.client import BatchAPIClient
from relevanceai.config import CONFIG
from relevanceai.errors import APIError

vis_requirements = False
try:
    from relevanceai.visualise.projector import Projector

    vis_requirements = True
except ModuleNotFoundError as e:
    pass

from relevanceai.vector_tools.client import VectorTools

def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class Client(BatchAPIClient, DocUtils):
    """Python Client for Relevance AI's relevanceai"""

    WELCOME_MESSAGE = """Welcome to the RelevanceAI Python SDK"""
    FAIL_MESSAGE = """Your API key is invalid. Please login again"""

    def __init__(
        self,
        project = os.getenv("VDB_PROJECT", None),
        api_key = os.getenv("VDB_API_KEY", None),
        verbose: bool=True

    ):

        if project is None or api_key is None:
            project, api_key = Client.token_to_auth(verbose=verbose)

        super().__init__(project, api_key)

        # if self.check_auth():
        #     if verbose: self.logger.success(self.WELCOME_MESSAGE)
        # else:
        #     raise APIError(self.FAIL_MESSAGE)

        if vis_requirements:
            self.projector = Projector(project, api_key)

        self.vector_tools = VectorTools(project, api_key)

    @property
    def output_format(self):
        return CONFIG.get_field("api.output_format", CONFIG.config)

    @output_format.setter
    def output_format(self, value):
        CONFIG.set_option("api.output_format", value)

    @staticmethod
    def token_to_auth(verbose=True):
        if verbose:
            print("You can sign up/login and find your credentials here: https://cloud.relevance.ai/sdk/api")
            print("Once you have signed up, click on the value under `Authorization token` and paste it here:")
        token = getpass.getpass("Authorization token (you can find it here: https://cloud.relevance.ai/sdk/api")
        project = token.split(":")[0]
        api_key = token.split(":")[1]
        os.environ["VDB_PROJECT"] = project
        os.environ["VDB_API_KEY"] = api_key
        return project, api_key

    @staticmethod
    def login(
        verbose: bool = True,
    ):
        """Preferred login method for demos and interactive usage."""
        project, api_key = Client.token_to_auth()
        return Client(
            project=project, api_key=api_key, verbose=verbose
        )

    @property
    def auth_header(self):
        return {"Authorization": self.project + ":" + self.api_key}

    def make_search_suggestion(self):
        return self.services.search.make_suggestion() 

    def check_auth(self):
        """TODO: Add a proper way to check authentication based on pinging.
        """
        response = self.datasets.list()
        try:
            return response.status_code == 200    
        except:
            raise Exception("Invalid auth details.")
